@@ -34,13 +34,12 @@
 
 #import "LUDictionary.h"
 #import "LUAgent.h"
-#import <NetInfo/dns.h>
-#import <NetInfo/sdns.h>
+#import <dns.h>
 
 @interface DNSAgent : LUAgent
 {
 	LUDictionary *stats;
-	sdns_handle_t *dns;
+	dns_handle_t dns;
 	BOOL allHostsEnabled;
 	BOOL localMulticast;
 	BOOL qualifyLocal;
@@ -48,5 +47,7 @@
 
 - (char **)searchList;
 
+- (LUDictionary *)dns_proxy:(LUDictionary *)pattern;
+
 @end
 
