@@ -44,12 +44,26 @@
 #import <sys/socket.h>
 #import <arpa/inet.h>
 #import <string.h>
+#import <resolv.h>
 #import <arpa/nameser.h>
 #import <sys/param.h>
 #import <NetInfo/dsutil.h>
 #import <stdio.h>
 #import <string.h>
 #import <libc.h>
+#import <dns_util.h>
+#import <ifaddrs.h>
+
+#define DNS_FLAGS_QR_MASK  0x8000
+#define DNS_FLAGS_QR_QUERY 0x0000
+#define DNS_FLAGS_AA 0x0400
+#define DNS_FLAGS_TC 0x0200
+#define DNS_FLAGS_RD 0x0100
+#define DNS_FLAGS_RA 0x0080
+
+#define DNS_FLAGS_OPCODE_MASK    0x7800
+
+#define DNS_FLAGS_RCODE_MASK 0x000f
 
 #define DefaultTimeout 30
 
@@ -63,52 +77,182 @@
 
 #define RFC_2782
 
+#define DNS_BUFFER_SIZE 8192
+#define DNS_BASE_64_BUFFER_SIZE 11432
+
+#define IP6_DNS_ZONE "ip6.arpa."
+
 extern char *nettoa(unsigned long);
+extern void __dns_open_notify();
 
 static const char hexchar[] = "0123456789abcdef";
-#ifdef _OS_NEXT_
-int
-inet_aton(char *a, struct in_addr *s)
-{
-	if (a == NULL) return 0;
-	if (s == NULL) return 0;
-	s->s_addr = inet_addr(a);
-	if (s->s_addr == -1) return 0;
-	return 1;
-}
-#endif	
 
-@implementation DNSAgent
+#ifdef DEBUG
+
+static char *_log_status_string(dns_reply_t *r)
+{
+	static char str[64];
+
+	if (r->status == DNS_STATUS_OK) return "OK";
+	if (r->status == DNS_STATUS_TIMEOUT) return "timeout";
+	if (r->status == DNS_STATUS_SEND_FAILED) return "send failed";
+	if (r->status == DNS_STATUS_RECEIVE_FAILED) return "receive failed";
+	snprintf(str, 64, "%u", r->status);
+	return str;
+}
+
+static char *_log_qr_string(dns_header_t *h)
+{
+	if ((h->flags & DNS_FLAGS_QR_MASK) == DNS_FLAGS_QR_QUERY) return "Q";
+	return "R";
+}
+
+static char *_log_opcode_string(dns_header_t *h)
+{
+	static char str[64];
+
+	switch (h->flags & DNS_FLAGS_OPCODE_MASK)
+	{
+		case ns_o_query:  return "std";
+		case ns_o_iquery: return "inv";
+		case ns_o_status: return "status";
+		case ns_o_notify: return "notify";
+		case ns_o_update: return "update";
+		default:
+		{
+			snprintf(str, 64, "%hu", (h->flags & DNS_FLAGS_OPCODE_MASK) >> 11);
+			return str;
+		}
+	}
+
+	return "???";
+}
+
+static char *_log_bool(u_int16_t b)
+{
+	if (b == 0) return "0";
+	return "1";
+}
+
+static char *_log_rcode_string(u_int16_t r)
+{
+	static char str[64];
+	if (r == ns_r_noerror)  return "OK";
+	if (r == ns_r_formerr)  return "format error";
+	if (r == ns_r_servfail) return "server failure";
+	if (r == ns_r_nxdomain) return "name error";
+	if (r == ns_r_notimpl)  return "not implemented";
+	if (r == ns_r_refused)  return "refused";
+
+	snprintf(str, 64, "%hu", r);
+	return str;
+}
+
+static char *_log_server_string(struct sockaddr *s)
+{
+	static char str[1024];
+	int offset;
+
+	offset = 4;
+	if (s->sa_family == AF_INET6) offset = 8;
+	inet_ntop(s->sa_family, (char *)(s) + offset, str, 1024);
+	return str;
+}
+
+void
+log_reply(dns_reply_t *r)
+{
+	dns_header_t *h;
+
+	if (r == NULL)
+	{
+		syslog(LOG_ERR, "*** NULL DNS REPLY ***");
+		return;
+	}
+
+	h = r->header;
+
+	syslog(LOG_ERR, "*** DNS: %s/%u/%s/%s/%s/%s%s%s%s/%s/%hu %hu %hu %hu",
+		_log_server_string(r->server), h->xid, _log_status_string(r), _log_qr_string(h), _log_opcode_string(h),
+		_log_bool(h->flags & DNS_FLAGS_AA), _log_bool(h->flags & DNS_FLAGS_TC), 
+		_log_bool(h->flags & DNS_FLAGS_RD), _log_bool(h->flags & DNS_FLAGS_RA),
+		_log_rcode_string(h->flags & DNS_FLAGS_RCODE_MASK),
+		h->qdcount, h->ancount, h->nscount, h->arcount);
+}
+
+void
+log_query(char *name, u_int32_t class, u_int32_t type)
+{
+	syslog(LOG_ERR, "??? DNS: %s %s %s", name, dns_class_string(class), dns_type_string(type));
+}
+#endif DEBUG
 
 int
 msg_callback(int priority, char *msg)
 {
+#ifdef DEBUG
+	priority = LOG_ERR;
+#endif
 	system_log(priority, msg);
 	return 0;
 }
 
 static char *
-_dns_inet_ntop(struct in6_addr a)
+_dns_inet_ntop(struct in6_addr a, u_int32_t iface)
 {
-	static char buf[128];
-	char t[32];
-	unsigned short x;
-	char *p;
-	int i;
+	char buf[128], scratch[128];
+	struct sockaddr_in6 s6, *if6;
+	void *p;
+	u_int16_t x, num;
+	struct ifaddrs *ifa, *ifp;
+
+	p = &s6;
+	p += 8;
 
 	memset(buf, 0, 128);
+	memset(&s6, 0, sizeof(struct sockaddr_in6));
+	memcpy(&(s6.sin6_addr), &a, sizeof(struct in6_addr));
 
-	p = (char *)&a.__u6_addr.__u6_addr32;
-	for (i = 0; i < 8; i++, x += 1)
+	if ((iface > 0) && (IN6_IS_ADDR_LINKLOCAL(&a) || IN6_IS_ADDR_SITELOCAL(&a)))
 	{
-		memmove(&x, p, 2);
-		p += 2;
-		sprintf(t, "%hx", x);
-		strcat(buf, t);
-		if (i < 7) strcat(buf, ":");
+		if (!strcmp("lo0", if_indextoname(iface, scratch)))
+		{
+			/*
+			 * Reply was over loopback.
+			 * If this is one of my addresses (as seen in getifaddrs)
+			 * then substitute the real interface number.
+			 */
+
+			if (getifaddrs(&ifa) < 0) ifa = NULL;
+
+			for (ifp = ifa; ifp != NULL; ifp = ifp->ifa_next)
+			{
+				if (ifp->ifa_addr == NULL) continue;
+				if ((ifp->ifa_flags & IFF_UP) == 0) continue;
+				if (ifp->ifa_addr->sa_family != AF_INET6) continue;
+
+				if6 = (struct sockaddr_in6 *)ifp->ifa_addr;
+
+				num = if6->sin6_addr.__u6_addr.__u6_addr16[1];
+				if6->sin6_addr.__u6_addr.__u6_addr16[1] = 0;
+
+				if (IN6_ARE_ADDR_EQUAL(&(if6->sin6_addr), &a))
+				{
+					iface = num;
+					break;
+				}
+			}
+
+			if (ifa != NULL) freeifaddrs(ifa);   
+		}
+
+		x = iface;
+		s6.sin6_addr.__u6_addr.__u6_addr16[1] = x;
 	}
 
-	return buf;
+	inet_ntop(AF_INET6, p, buf, 128);
+
+	return copyString(buf);
 }
 
 static char *
@@ -120,7 +264,7 @@ reverse_ipv4(char *v4)
 		unsigned char b[4];
 	} ab;
 	struct in_addr s;
-	char name[32], *p;
+	char *p;
 
 	if (v4 == NULL) return NULL;
 
@@ -128,10 +272,7 @@ reverse_ipv4(char *v4)
 	
 	ab.a = ntohl(s.s_addr);
 
-	sprintf(name, "%u.%u.%u.%u.in-addr.arpa.",
-		ab.b[3], ab.b[2], ab.b[1], ab.b[0]);
-
-	p = copyString(name);
+	asprintf(&p, "%u.%u.%u.%u.in-addr.arpa.", ab.b[3], ab.b[2], ab.b[1], ab.b[0]);
 	return p;
 }
 
@@ -141,7 +282,7 @@ reverse_ipv6(char *v6)
 	char x[65], *p;
 	int i, j;
 	u_int8_t d, hi, lo;
-        struct in6_addr a6;
+	struct in6_addr a6;
 
 	if (v6 == NULL) return NULL;
 	if (inet_pton(AF_INET6, v6, &a6) == 0) return NULL;
@@ -161,149 +302,16 @@ reverse_ipv6(char *v6)
 
 	p = calloc(1, 72);
 	memmove(p, x, 64);
-	strcat(p, "ip6.int");
+	strcat(p, IP6_DNS_ZONE);
 
 	return p;
 }
 
-+ (DNSAgent *)alloc
-{
-	DNSAgent *agent;
-
-	agent = [super alloc];
-	system_log(LOG_DEBUG, "Allocated DNSAgent 0x%08x\n", (int)agent);
-	return agent;
-}
-
-+ (int)typeForKey:(char *)key
-{
-	if (key == NULL) return DNS_TYPE_ANY;
-
-	if (streq(key, "ip_address")) return DNS_TYPE_A;
-	if (streq(key, "nameserver")) return DNS_TYPE_NS;
-	if (streq(key, "name")) return DNS_TYPE_CNAME;
-	if (streq(key, "authority")) return DNS_TYPE_SOA;
-	if (streq(key, "MB")) return DNS_TYPE_MB;
-	if (streq(key, "MG")) return DNS_TYPE_MG;
-	if (streq(key, "MR")) return DNS_TYPE_MR;
-	if (streq(key, "NULL")) return DNS_TYPE_NULL;
-	if (streq(key, "WKS")) return DNS_TYPE_WKS;
-	if (streq(key, "PTR")) return DNS_TYPE_PTR;
-	if (streq(key, "hostinfo")) return DNS_TYPE_HINFO;
-	if (streq(key, "MINFO")) return DNS_TYPE_MINFO;
-	if (streq(key, "mail_exchanger")) return DNS_TYPE_MX;
-	if (streq(key, "TXT")) return DNS_TYPE_TXT;
-	if (streq(key, "RP")) return DNS_TYPE_RP;
-	if (streq(key, "AFSDB")) return DNS_TYPE_AFSDB;
-	if (streq(key, "X25")) return DNS_TYPE_X25;
-	if (streq(key, "isdn")) return DNS_TYPE_ISDN;
-	if (streq(key, "RT")) return DNS_TYPE_RT;
-	if (streq(key, "ipv6_address")) return DNS_TYPE_AAAA;
-	if (streq(key, "location")) return DNS_TYPE_LOC;
-	if (streq(key, "service")) return DNS_TYPE_SRV;
-	if (streq(key, "AXFR")) return DNS_TYPE_AXFR;
-	if (streq(key, "service")) return DNS_TYPE_SRV;
-	if (streq(key, "MAILB")) return DNS_TYPE_MAILB;
-	if (streq(key, "MAILA")) return DNS_TYPE_MAILA;
-
-	return -1;
-}
-
-+ (char *)keyForType:(int)t
-{
-	switch (t)
-	{
-		case DNS_TYPE_ANY: return "any";
-		case DNS_TYPE_A: return "ip_address";
-		case DNS_TYPE_NS: return "nameserver";
-		case DNS_TYPE_CNAME: return "name";
-		case DNS_TYPE_SOA: return "authority";
-		case DNS_TYPE_MB: return "MB";
-		case DNS_TYPE_MG: return "MG";
-		case DNS_TYPE_MR: return "MR";
-		case DNS_TYPE_NULL: return "NULL";
-		case DNS_TYPE_WKS: return "WKS";
-		case DNS_TYPE_PTR: return "name";
-		case DNS_TYPE_HINFO: return "hostinfo";
-		case DNS_TYPE_MINFO: return "MINFO";
-		case DNS_TYPE_MX: return "mail_exchanger";
-		case DNS_TYPE_TXT: return "TXT";
-		case DNS_TYPE_RP: return "RP";
-		case DNS_TYPE_AFSDB: return "AFSDB";
-		case DNS_TYPE_X25: return "X25";
-		case DNS_TYPE_ISDN: return "isdn";
-		case DNS_TYPE_RT: return "RT";
-		case DNS_TYPE_AAAA: return "ipv6_address";
-		case DNS_TYPE_LOC: return "location";
-		case DNS_TYPE_SRV: return "service";
-		case DNS_TYPE_AXFR: return "AXFR";
-		case DNS_TYPE_MAILB: return "MAILB";
-		case DNS_TYPE_MAILA: return "MAILA";
-		default: return NULL;
-	}
-	return NULL;
-}
-
-#ifdef NOTDEF
-static char *
-type_string(u_int16_t t)
-{
-	static char str[32];
-
-	switch (t)
-	{
-		case DNS_TYPE_A:     return "A";
-		case DNS_TYPE_AAAA:  return "AAAA";
-		case DNS_TYPE_NS:    return "NS";
-		case DNS_TYPE_CNAME: return "CNAME";
-		case DNS_TYPE_SOA:   return "SOA";
-		case DNS_TYPE_MB:    return "MB";
-		case DNS_TYPE_MG:    return "MG";
-		case DNS_TYPE_MR:    return "MR";
-		case DNS_TYPE_NULL:  return "NULL";
-		case DNS_TYPE_WKS:   return "WKS";
-		case DNS_TYPE_PTR:   return "PTR";
-		case DNS_TYPE_HINFO: return "HINFO";
-		case DNS_TYPE_MINFO: return "MINFO";
-		case DNS_TYPE_MX:    return "MX";
-		case DNS_TYPE_TXT:   return "TXT";
-		case DNS_TYPE_RP:    return "PR";
-		case DNS_TYPE_AFSDB: return "AFSDB";
-		case DNS_TYPE_X25:   return "X25";
-		case DNS_TYPE_ISDN:  return "ISDN";
-		case DNS_TYPE_RT:    return "RT";
-		case DNS_TYPE_LOC:   return "LOC";
-		case DNS_TYPE_SRV:   return "SRV";
-		case DNS_TYPE_ANY:   return "ANY";
-	}
-
-	sprintf(str, "%5hu", t);
-	return str;
-}
-
-static char *
-class_string(u_int16_t c)
-{
-	static char str[32];
-
-	switch (c)
-	{
-		case DNS_CLASS_IN: return "IN";
-		case DNS_CLASS_CS: return "CS";
-		case DNS_CLASS_CH: return "CH";
-		case DNS_CLASS_HS: return "HS";
-	}
-
-	sprintf(str, "%2hu", c);
-	return str;
-}
-#endif
-
 static char *
 coord_ntoa(int32_t coord, u_int32_t islat)
 {
 	int32_t deg, min, sec, secfrac;
-	static char buf[64];
+	char *p;
 	char dir;
 
 	coord = coord - 0x80000000;
@@ -333,15 +341,15 @@ coord_ntoa(int32_t coord, u_int32_t islat)
 	coord = coord / 60;
 	deg = coord;
 
-	sprintf(buf, "%d %.2d %.2d.%.3d %c", deg, min, sec, secfrac, dir);
-	return buf;
+	asprintf(&p, "%d %.2d %.2d.%.3d %c", deg, min, sec, secfrac, dir);
+	return p;
 }
 
 static char *
 alt_ntoa(int32_t alt)
 {
 	int32_t ref, m, frac, sign;
-	static char buf[128];
+	char *p;
 
 	ref = 100000 * 100;
 	sign = 1;
@@ -359,8 +367,8 @@ alt_ntoa(int32_t alt)
 	frac = alt % 100;
 	m = (alt / 100) * sign;
 	
-	sprintf(buf, "%d.%.2d", m, frac);
-	return buf;
+	asprintf(&p, "%d.%.2d", m, frac);
+	return p;
 }
 
 static unsigned int poweroften[10] =
@@ -379,7 +387,7 @@ static unsigned int poweroften[10] =
 static char *
 precsize_ntoa(u_int8_t prec)
 {
-	static char buf[19];
+	char *p;
 	unsigned long val;
 	int mantissa, exponent;
 
@@ -388,61 +396,38 @@ precsize_ntoa(u_int8_t prec)
 
 	val = mantissa * poweroften[exponent];
 
-	sprintf(buf, "%ld.%.2ld", val/100, val%100);
-	return buf;
+	asprintf(&p, "%ld.%.2ld", val/100, val%100);
+	return p;
+}
+
+@implementation DNSAgent
+
++ (DNSAgent *)alloc
+{
+	DNSAgent *agent;
+
+	agent = [super alloc];
+	system_log(LOG_DEBUG, "Allocated DNSAgent 0x%08x\n", (int)agent);
+	return agent;
 }
 
 - (LUAgent *)initWithArg:(char *)arg
 {
-	LUDictionary *config;
-	char *dn;
-	struct timeval t;
-	unsigned long r;
-	char str[256];
-
 	[super initWithArg:arg];
 
-	t.tv_sec = 0; 
-	t.tv_usec = 0;
-
-	config = [configManager configGlobal:configurationArray];
-	t.tv_sec = [configManager intForKey:"Timeout" dict:config default:DefaultTimeout];
-
-	config = [configManager configForAgent:"DNSAgent" fromConfig:configurationArray];
-	t.tv_sec = [configManager intForKey:"Timeout" dict:config default:t.tv_sec];
-
-	dn = [configManager stringForKey:"Domain" dict:config default:NULL];
-	r = [configManager intForKey:"Retries" dict:config default:2];
-	allHostsEnabled =  [configManager boolForKey:"AllHostsEnabled" dict:config default:NO];
-	qualifyLocal =  [configManager boolForKey:"QualifyLocalDomain" dict:config default:NO];
-
 	syslock_lock(rpcLock);
-	/* SDNS API */
-	dns = sdns_open();
+	/* DNS API */
+	__dns_open_notify();
+	dns = dns_open(NULL);
 	syslock_unlock(rpcLock);
 
-	freeString(dn);
-
 	if (dns == NULL)
 	{
 		[self release];
 		return nil;
 	}
 
-	if (t.tv_sec == 0) t.tv_sec = 1;
-
-	/* SDNS API */
-	sdns_open_log(dns, "DNSAgent", DNS_LOG_CALLBACK, NULL, 0, 0, msg_callback);
-	sprintf(str, "DNSAgent");
-
-#ifdef NOTDEF
-/* XXX NO SDNS SUPPORT XXX */
-	dns_set_server_retries(dns, r);
-	dns_set_timeout(dns, &t);
-#endif
-
-	[self setBanner:str];
-
+	[self setBanner:"DNSAgent"];
 	return self;
 }
 
@@ -459,8 +444,8 @@ precsize_ntoa(u_int8_t prec)
 - (void)dealloc
 {
 	if (stats != nil) [stats release];
-	/* SDNS API */
-	sdns_free(dns);
+	/* DNS API */
+	dns_free(dns);
 	system_log(LOG_DEBUG, "Deallocated DNSAgent 0x%08x\n", (int)self);
 	[super dealloc];
 }
@@ -490,86 +475,113 @@ precsize_ntoa(u_int8_t prec)
 	return item;
 }
 
-- (LUDictionary *)dictForDNSResouceRecord:(dns_resource_record_t *)r
+- (LUDictionary *)dictForDNSResouceRecord:(dns_resource_record_t *)r interface:(unsigned int)iface
 {
 	LUDictionary *item;
-	char str[32];
+	char str[32], *ifname, *addrstr, *x;
 	int i;
 
 	if (r == NULL) return nil;
 
-	item = [[LUDictionary alloc] init];
+	item = [[LUDictionary alloc] initTimeStamped];
 
-#ifdef NOTDEF
-	[item setValue:type_string(r->type) forKey:"DNS_type"];
-	[item setValue:class_string(r->class) forKey:"DNS_class"];
-	sprintf(str, "%u", r->ttl);
-	[item setValue:str forKey:"DNS_ttl"];
-#endif
-
-	switch (r->type)
+	switch (r->dnstype)
 	{
-		case DNS_TYPE_A:
+		case ns_t_a:
 			[item setValue:r->name forKey:"name"];
 			[item setValue:inet_ntoa(r->data.A->addr) forKey:"ip_address"];
+			if (iface > 0)
+			{
+				asprintf(&ifname, "%u", iface);
+				[item setValue:ifname forKey:"interface"];
+				free(ifname);
+			}
 			break;
 
-		case DNS_TYPE_AAAA:
+		case ns_t_aaaa:
 			[item setValue:r->name forKey:"name"];
-			[item setValue:_dns_inet_ntop(r->data.AAAA->addr) forKey:"ipv6_address"];
-			break;
 
-		case DNS_TYPE_CNAME:
-			[item setValue:r->name forKey:"name"];
+			addrstr = _dns_inet_ntop(r->data.AAAA->addr, iface);
+			[item setValue:addrstr forKey:"ipv6_address"];
+			if (addrstr != NULL) free(addrstr);
+
+			if (iface > 0)
+			{
+				asprintf(&ifname, "%u", iface);
+				[item setValue:ifname forKey:"interface"];
+				free(ifname);
+			}
+		break;
+
+		case ns_t_cname:
+			[item setValue:r->name forKey:"alias"];
 			[item setValue:r->data.CNAME->name forKey:"cname"];
 			break;
 
-		case DNS_TYPE_MB:
-		case DNS_TYPE_MG:
-		case DNS_TYPE_MR:
-		case DNS_TYPE_PTR:
-		case DNS_TYPE_NS:
-			[item setValue:r->data.CNAME->name forKey:"name"];
+		case ns_t_mb:
+			[item setValue:r->name forKey:"user"];
+			[item setValue:r->data.CNAME->name forKey:"mailhost"];
 			break;
 
-		case DNS_TYPE_SOA:
+		case ns_t_mg:
+			[item setValue:r->name forKey:"mailgroup"];
+			[item setValue:r->data.CNAME->name forKey:"member"];
+			break;
+
+		case ns_t_mr:
+			[item setValue:r->name forKey:"alias"];
+			[item setValue:r->data.CNAME->name forKey:"user"];
+			break;
+
+		case ns_t_ptr:
+			[item setValue:r->name forKey:"name"];
+			[item setValue:r->data.CNAME->name forKey:"ptrdname"];
+			break;
+
+		case ns_t_ns:
+			[item setValue:r->name forKey:"domain"];
+			[item setValue:r->data.CNAME->name forKey:"server"];
+			break;
+
+		case ns_t_soa:
 			[item setValue:r->name forKey:"name"];
 			[item setValue:r->data.SOA->mname forKey:"mname"];
 			[item setValue:r->data.SOA->rname forKey:"rname"];
-			sprintf(str, "%u", r->data.SOA->serial);
+			snprintf(str, 32, "%u", r->data.SOA->serial);
 			[item setValue:str forKey:"serial"];
-			sprintf(str, "%u", r->data.SOA->refresh);
+			snprintf(str, 32, "%u", r->data.SOA->refresh);
 			[item setValue:str forKey:"refresh"];
-			sprintf(str, "%u", r->data.SOA->retry);
+			snprintf(str, 32, "%u", r->data.SOA->retry);
 			[item setValue:str forKey:"retry"];
-			sprintf(str, "%u", r->data.SOA->expire);
+			snprintf(str, 32, "%u", r->data.SOA->expire);
 			[item setValue:str forKey:"expire"];
-			sprintf(str, "%u", r->data.SOA->minimum);
+			snprintf(str, 32, "%u", r->data.SOA->minimum);
 			[item setValue:str forKey:"minimum"];
 			break;
 
-		case DNS_TYPE_WKS:
+		case ns_t_wks:
 			break;
 
-		case DNS_TYPE_HINFO:
+		case ns_t_hinfo:
 			[item setValue:r->name forKey:"name"];
 			[item setValue:r->data.HINFO->cpu forKey:"cpu"];
 			[item setValue:r->data.HINFO->os forKey:"os"];
 			break;
 
-		case DNS_TYPE_MINFO:
+		case ns_t_minfo:
+			[item setValue:r->name forKey:"name"];
 			[item setValue:r->data.MINFO->rmailbx forKey:"rmailbx"];
 			[item setValue:r->data.MINFO->emailbx forKey:"emailbx"];
 			break;
 
-		case DNS_TYPE_MX:
+		case ns_t_mx:
 			[item setValue:r->name forKey:"name"];
-			sprintf(str, "%u", r->data.MX->preference);
+			snprintf(str, 32, "%u", r->data.MX->preference);
 			[item setValue:str forKey:"preference"];
 			[item setValue:r->data.MX->name forKey:"mail_exchanger"];
 			break;
 			
-		case DNS_TYPE_TXT:
+		case ns_t_txt:
 			[item setValue:r->name forKey:"name"];
 			for (i=0; i<r->data.TXT->string_count; i++)
 			{
@@ -577,49 +589,67 @@ precsize_ntoa(u_int8_t prec)
 			}
 			break;
 
-		case DNS_TYPE_RP:
+		case ns_t_rp:
 			[item setValue:r->data.RP->mailbox forKey:"mailbox"];
 			[item setValue:r->data.RP->txtdname forKey:"txtdname"];
 			break;
 
-		case DNS_TYPE_AFSDB:
-			sprintf(str, "%u", r->data.AFSDB->subtype);
+		case ns_t_afsdb:
+			snprintf(str, 32, "%u", r->data.AFSDB->subtype);
 			[item setValue:str forKey:"subtype"];
 			[item setValue:r->data.AFSDB->hostname forKey:"name"];
 			break;
 
-		case DNS_TYPE_X25:
+		case ns_t_x25:
 			[item setValue:r->data.X25->psdn_address forKey:"psdn_address"];
 			break;
 
-		case DNS_TYPE_ISDN:
+		case ns_t_isdn:
 			[item setValue:r->data.ISDN->isdn_address forKey:"isdn_address"];
 			if (r->data.ISDN->subaddress != NULL)
 				[item setValue:r->data.ISDN->subaddress forKey:"subaddress"];
 			break;
 
-		case DNS_TYPE_RT:
-			sprintf(str, "%hu", r->data.RT->preference);
+		case ns_t_rt:
+			snprintf(str, 32, "%hu", r->data.RT->preference);
 			[item setValue:str forKey:"preference"];
 			[item setValue:r->data.RT->intermediate forKey:"intermediate"];
 			break;
 
-		case DNS_TYPE_LOC:
-			[item setValue:coord_ntoa(r->data.LOC->latitude, 1) forKey:"latitude"];
-			[item setValue:coord_ntoa(r->data.LOC->longitude, 1) forKey:"longitude"];
-			[item setValue:alt_ntoa(r->data.LOC->altitude) forKey:"altitude"];
-			[item setValue:precsize_ntoa(r->data.LOC->size) forKey:"size"];
-			[item setValue:precsize_ntoa(r->data.LOC->horizontal_precision) forKey:"horizontal_precision"];
-			[item setValue:precsize_ntoa(r->data.LOC->vertical_precision) forKey:"vertical_precision"];
+		case ns_t_loc:
+			x = coord_ntoa(r->data.LOC->latitude, 1);
+			[item setValue:x forKey:"latitude"];
+			if (x != NULL) free(x);
+
+			x = coord_ntoa(r->data.LOC->longitude, 1);
+			[item setValue:x forKey:"longitude"];
+			if (x != NULL) free(x);
+
+			x = alt_ntoa(r->data.LOC->altitude);
+			[item setValue:x forKey:"altitude"];
+			if (x != NULL) free(x);
+
+			x = precsize_ntoa(r->data.LOC->size);
+			[item setValue:x forKey:"size"];
+			if (x != NULL) free(x);
+
+			x = precsize_ntoa(r->data.LOC->horizontal_precision);
+			[item setValue:x forKey:"horizontal_precision"];
+			if (x != NULL) free(x);
+
+			x = precsize_ntoa(r->data.LOC->vertical_precision);
+			[item setValue:x forKey:"vertical_precision"];
+			if (x != NULL) free(x);
+
 			break;
 
-		case DNS_TYPE_SRV:
+		case ns_t_srv:
 			[item setValue:r->name forKey:"name"];
-			sprintf(str, "%hu", r->data.SRV->priority);
+			snprintf(str, 32, "%hu", r->data.SRV->priority);
 			[item setValue:str forKey:"priority"];
-			sprintf(str, "%hu", r->data.SRV->weight);
+			snprintf(str, 32, "%hu", r->data.SRV->weight);
 			[item setValue:str forKey:"weight"];
-			sprintf(str, "%hu", r->data.SRV->port);
+			snprintf(str, 32, "%hu", r->data.SRV->port);
 			[item setValue:str forKey:"port"];
 			[item setValue:r->data.SRV->target forKey:"target"];
 			break;
@@ -632,54 +662,84 @@ precsize_ntoa(u_int8_t prec)
 {
 	LUDictionary *item;
 	LUArray *list;
-	int i;
+	int i, iface;
 
 	if (r == NULL) return nil;
 	if (r->status != DNS_STATUS_OK) return nil;
-	if ((r->header->flags & DNS_FLAGS_RCODE_MASK) != DNS_FLAGS_RCODE_NO_ERROR)
+	if ((r->header->flags & DNS_FLAGS_RCODE_MASK) != ns_r_noerror)
 		return nil;
 
 	if (r->header->ancount == 0) return nil;
 
 	list = [[LUArray alloc] init];
 
+	iface = 0;
+	if (r->server->sa_family == AF_INET)
+	{
+		memcpy(&iface, (((struct sockaddr_in *)(r->server))->sin_zero), 4);
+	}
+	else if (r->server->sa_family == AF_INET6)
+	{
+		iface = ((struct sockaddr_in6 *)(r->server))->sin6_scope_id;
+	}
+
  	for (i = 0; i < r->header->ancount; i++)
 	{
-		item = [self dictForDNSResouceRecord:r->answer[i]];
+		item = [self dictForDNSResouceRecord:r->answer[i] interface:iface];
 		if  (item != nil)
 		{
 			[list addObject:item];
 			[item release];
 		}
 	}
-	
+
 	return list;
 }
 
 - (LUDictionary *)dictForDNSReply:(dns_reply_t *)r
 {
 	LUDictionary *host;
-	char *longName = NULL;
-	char *shortName = NULL;
+	char *tempName = NULL;
+	char *realName = NULL;
 	char *domainName = NULL;
-	int i, got_data, ttl;
+	int i, got_data, ttl, offset, iface;
 	time_t now;
-	char scratch[32];
-	int alias;
+	char scratch[256], *addrstr;
+	int name_count, cname_count, alias_count;
 
 	if (r == NULL) return nil;
 	if (r->status != DNS_STATUS_OK) return nil;
-	if ((r->header->flags & DNS_FLAGS_RCODE_MASK) != DNS_FLAGS_RCODE_NO_ERROR)
+	if ((r->header->flags & DNS_FLAGS_RCODE_MASK) != ns_r_noerror)
 		return nil;
 
 	if (r->header->ancount == 0) return nil;
 
- 	host = [[LUDictionary alloc] init];
-	[host setValue:inet_ntoa(r->server) forKey:"_lookup_DNS_server"];
+ 	host = [[LUDictionary alloc] initTimeStamped];
 
-	alias = 0;
+	offset = 4;
+	if (r->server->sa_family == AF_INET6) offset = 8;
+
+	scratch[0] = '\0';
+	inet_ntop(r->server->sa_family, (char *)(r->server) + offset, scratch, 256);
+	if (scratch[0] != '\0') [host setValue:scratch forKey:"_lookup_DNS_server"];
+
+	iface = 0;
+	if (r->server->sa_family == AF_INET)
+	{
+		memcpy(&iface, (((struct sockaddr_in *)(r->server))->sin_zero), 4);
+	}
+	else if (r->server->sa_family == AF_INET6)
+	{
+		iface = ((struct sockaddr_in6 *)(r->server))->sin6_scope_id;
+	}
+
+	snprintf(scratch, 256, "%u", iface);
+	[host mergeValue:scratch forKey:"interface"];
+
+	name_count = 0;
+	cname_count = 0;
+	alias_count = 0;
 	got_data = 0;
-	longName = lowerCase(r->answer[0]->name);
 
 	ttl = r->answer[0]->ttl;
 
@@ -687,43 +747,47 @@ precsize_ntoa(u_int8_t prec)
 	{
 		if (r->answer[i]->ttl < ttl) ttl = r->answer[i]->ttl;
 
-		if (r->answer[i]->type == DNS_TYPE_A)
+		if (r->answer[i]->dnstype == ns_t_a)
 		{
 			got_data++;
-			[host mergeValue:inet_ntoa(r->answer[i]->data.A->addr)
-				forKey:"ip_address"];
+			[host mergeValue:inet_ntoa(r->answer[i]->data.A->addr) forKey:"ip_address"];
 		}
 
-		else if (r->answer[i]->type == DNS_TYPE_AAAA)
+		else if (r->answer[i]->dnstype == ns_t_aaaa)
 		{
 			got_data++;
-			[host mergeValue:_dns_inet_ntop(r->answer[i]->data.AAAA->addr)
-				forKey:"ipv6_address"];
+			addrstr = _dns_inet_ntop(r->answer[i]->data.AAAA->addr, iface);
+			[host mergeValue:addrstr forKey:"ipv6_address"];
+			if (addrstr != NULL) free(addrstr);
 		}
 
-		else if (r->answer[i]->type == DNS_TYPE_CNAME)
+		else if (r->answer[i]->dnstype == ns_t_cname)
 		{
 			got_data++;
-			alias++;
-			freeString(longName);
-			longName = lowerCase(r->answer[i]->name);
-			[host mergeValue:longName forKey:"alias"];
 
-			/* Real name of the host is value of CNAME */
-			freeString(longName);
-			longName = lowerCase(r->answer[i]->data.CNAME->name);
+			cname_count++;
+			tempName = lowerCase(r->answer[i]->data.CNAME->name);
+			[host mergeValue:tempName forKey:"cname"];
+			freeString(tempName);
+
+			alias_count++;
+			tempName = lowerCase(r->answer[i]->name);
+			[host mergeValue:tempName forKey:"alias"];
+			freeString(tempName);
 		}
 
-		else if (r->answer[i]->type == DNS_TYPE_PTR)
+		else if (r->answer[i]->dnstype == ns_t_ptr)
 		{
 			got_data++;
-			
+			name_count++;
+			tempName = lowerCase(r->answer[i]->data.PTR->name);
+			[host mergeValue:tempName forKey:"name"];
+			freeString(tempName);
+
 			/* Save referring name in case someone wants it later */
 			[host mergeValue:r->answer[i]->name forKey:"ptr_name"];
 
-			/* Real name of the host is value of PTR */
-			freeString(longName);
-			longName = lowerCase(r->answer[i]->data.PTR->name);
+			if (realName == NULL) realName = lowerCase(r->answer[i]->data.PTR->name);
 		}
 	}
 
@@ -733,32 +797,41 @@ precsize_ntoa(u_int8_t prec)
 		return nil;
 	}
 
-	sprintf(scratch, "%u", ttl);
+	if (realName == NULL)
+	{
+		/*
+		 * No PTR records.
+		 * If there was a CNAME record, use it
+		 * If not, use answer[0]->name
+		 */
+		if (cname_count > 0) realName = lowerCase([host valueForKey:"cname"]);
+		else realName = lowerCase(r->answer[0]->name);
+		
+		[host mergeValue:realName forKey:"name"];
+	}
+
+	snprintf(scratch, 256, "%u", ttl);
 	[host setValue:scratch forKey:"_lookup_DNS_time_to_live"];
 	now = time(0);
-	sprintf(scratch, "%lu", now);
+	snprintf(scratch, 256, "%lu", now);
 	[host setValue:scratch forKey:"_lookup_DNS_timestamp"];
 
-	[host mergeValue:longName forKey:"name"];
-
-	shortName = prefix(longName, '.');
-	domainName = postfix(longName, '.');
-
-	if (domainName != NULL)
+	if (realName != NULL)
 	{
-		[host setValue:domainName forKey:"_lookup_DNS_domain"];
-#ifdef NOTDEF
-/* XXX NO SDNS SUPPORT XXX */
-		if (strcmp(domainName, dns->domain) == 0)
-			[host mergeValue:shortName forKey:"name"];
-#endif
-	}
+		domainName = postfix(realName, '.');
+		if (domainName != NULL) [host setValue:domainName forKey:"_lookup_DNS_domain"];
 	
-	freeString(longName);
-	freeString(shortName);
-	freeString(domainName);
+		freeString(realName);
+		freeString(domainName);
+	}
 
-	if (alias > 0)
+	if (cname_count > 0)
+	{
+		[host mergeValues:[host valuesForKey:"cname"] forKey:"name"];
+		[host removeKey:"cname"];
+	}
+
+	if (alias_count > 0)
 	{
 		[host mergeValues:[host valuesForKey:"alias"] forKey:"name"];
 		[host removeKey:"alias"];
@@ -771,16 +844,17 @@ precsize_ntoa(u_int8_t prec)
 {
 	LUDictionary *host;
 	dns_reply_t *r;
-	dns_question_t q;
 
 	if (k == NULL) return nil;
 
-	q.class = DNS_CLASS_IN;	
-	q.type = t;
-	q.name = k;
-
-	/* SDNS API */
-	r = sdns_query(dns, &q);
+	/* DNS API */
+#ifdef DEBUG
+	log_query(k, ns_c_in, t);
+#endif
+	r = dns_lookup(dns, k, ns_c_in, t);
+#ifdef DEBUG
+	log_reply(r);
+#endif
 	host = [self dictForDNSReply:r];
 	dns_free_reply(r);
 
@@ -791,8 +865,7 @@ precsize_ntoa(u_int8_t prec)
 {
 	LUDictionary *host;
 	dns_reply_t *r;
-	dns_question_t q;
-	char name[32];
+	char name[64];
 	union
 	{
 		unsigned long a;
@@ -801,15 +874,17 @@ precsize_ntoa(u_int8_t prec)
 
 	ab.a = addr->s_addr;
 
-	sprintf(name, "%u.%u.%u.%u.in-addr.arpa.",
+	snprintf(name, 64, "%u.%u.%u.%u.in-addr.arpa.",
 		ab.b[3], ab.b[2], ab.b[1], ab.b[0]);
 
-	q.class = DNS_CLASS_IN;	
-	q.type = DNS_TYPE_PTR;
-	q.name = name;
-
-	/* SDNS API */
-	r = sdns_query(dns, &q);
+	/* DNS API */
+#ifdef DEBUG
+	log_query(name, ns_c_in, ns_t_ptr);
+#endif
+	r = dns_lookup(dns, name, ns_c_in, ns_t_ptr);
+#ifdef DEBUG
+	log_reply(r);
+#endif
 	host = [self dictForDNSReply:r];
 	dns_free_reply(r);
 
@@ -822,17 +897,18 @@ precsize_ntoa(u_int8_t prec)
 {
 	LUDictionary *net;
 	dns_reply_t *r;
-	dns_question_t q;
 	char **l;
 	char *longName, *shortName, *domainName;
-	char str[32];
+	char str[64];
 
-	q.class = DNS_CLASS_IN;	
-	q.type = DNS_TYPE_PTR;
-	q.name = name;
-
-	/* SDNS API */
-	r = sdns_query(dns, &q);
+	/* DNS API */
+#ifdef DEBUG
+	log_query(name, ns_c_in, ns_t_ptr);
+#endif
+	r = dns_lookup(dns, name, ns_c_in, ns_t_ptr);
+#ifdef DEBUG
+	log_reply(r);
+#endif
 	net = [self dictForDNSReply:r];
 	dns_free_reply(r);
 
@@ -852,7 +928,7 @@ precsize_ntoa(u_int8_t prec)
 		return nil;
 	}
 
-	sprintf(str, "%s.%s.%s.%s", l[3], l[2], l[1], l[0]);
+	snprintf(str, 64, "%s.%s.%s.%s", l[3], l[2], l[1], l[0]);
 	freeList(l);
 	
 	longName = [net valueForKey:"ptr_name"];
@@ -877,8 +953,7 @@ precsize_ntoa(u_int8_t prec)
 {
 	LUDictionary *net;
 	dns_reply_t *r;
-	dns_question_t q;
-	char name[32];
+	char name[64];
 	union
 	{
 		unsigned long a;
@@ -887,15 +962,17 @@ precsize_ntoa(u_int8_t prec)
 
 	ab.a = addr->s_addr;
 
-	sprintf(name, "%u.%u.%u.%u.in-addr.arpa.",
+	snprintf(name, 64, "%u.%u.%u.%u.in-addr.arpa.",
 		ab.b[3], ab.b[2], ab.b[1], ab.b[0]);
 
-	q.class = DNS_CLASS_IN;	
-	q.type = DNS_TYPE_PTR;
-	q.name = name;
-
-	/* SDNS API */
-	r = sdns_query(dns, &q);
+	/* DNS API */
+#ifdef DEBUG
+	log_query(name, ns_c_in, ns_t_ptr);
+#endif
+	r = dns_lookup(dns, name, ns_c_in, ns_t_ptr);
+#ifdef DEBUG
+	log_reply(r);
+#endif
 	net = [self dictForDNSReply:r];
 	dns_free_reply(r);
 
@@ -907,11 +984,17 @@ precsize_ntoa(u_int8_t prec)
 	return [self stamp:net];
 }
 
+- (char **)searchList
+{
+	return NULL;
+}
+
+#ifdef NOTDEF
 - (char **)searchList
 {
 	char **l, *dn, *s, *p;
-	int i;
-	dns_handle_t *h;
+	int i, count;
+	void *h;
 
 	h = dns_open(NULL);
 	if (h == NULL) return NULL;
@@ -946,13 +1029,14 @@ precsize_ntoa(u_int8_t prec)
 	dns_free(h);
 	return l;
 }
+#endif
 
 - (LUDictionary *)itemWithKey:(char *)key
 	value:(char *)val
 	category:(LUCategory)cat
 {
 	struct in_addr ip;
-	char *str, scratch[32];
+	char *str, scratch[64];
 	LUDictionary *item, *item2;
 	unsigned long ttl1, ttl2;
 
@@ -964,17 +1048,19 @@ precsize_ntoa(u_int8_t prec)
 		case LUCategoryHost:
 			if (streq(key, "name"))
 			{
-				item = [self hostWithKey:val dnstype:DNS_TYPE_A];
+				item = [self hostWithKey:val dnstype:ns_t_a];
 				if (item != nil) [item mergeValue:val forKey:"name"];
 				return item;
 			}
 			if (streq(key, "namev6"))
 			{
-				item = [self hostWithKey:val dnstype:DNS_TYPE_A];
+				item = [self hostWithKey:val dnstype:ns_t_a];
 				[item mergeValue:val forKey:"name"];
 				ttl1 = [item unsignedLongForKey:"_lookup_DNS_time_to_live"];
 
-				item2 = [self hostWithKey:val dnstype:DNS_TYPE_AAAA];
+				item2 = [self hostWithKey:val dnstype:ns_t_aaaa];
+				if (item == nil) return item2;
+
 				if (item2 != nil)
 				{
 					[item mergeKey:"name" from:item2];
@@ -984,7 +1070,7 @@ precsize_ntoa(u_int8_t prec)
 
 					if (ttl2 < ttl1)
 					{
-						sprintf(scratch, "%lu", ttl2);
+						snprintf(scratch, 64, "%lu", ttl2);
 						[item setValue:scratch forKey:"_lookup_DNS_time_to_live"];
 					}
 				}
@@ -994,7 +1080,7 @@ precsize_ntoa(u_int8_t prec)
 			{
 				str = reverse_ipv4(val);
 				if (str == NULL) return nil;
-				item = [self hostWithKey:str dnstype:DNS_TYPE_PTR];
+				item = [self hostWithKey:str dnstype:ns_t_ptr];
 				free(str);
 				if (item != nil) [item mergeValue:val forKey:"ip_address"];
 				return item;
@@ -1003,7 +1089,7 @@ precsize_ntoa(u_int8_t prec)
 			{
 				str = reverse_ipv6(val);
 				if (str == NULL) return nil;
-				item = [self hostWithKey:str dnstype:DNS_TYPE_PTR];
+				item = [self hostWithKey:str dnstype:ns_t_ptr];
 				free(str);
 				if (item != nil) [item mergeValue:val forKey:"ipv6_address"];
 				return item;
@@ -1024,20 +1110,50 @@ precsize_ntoa(u_int8_t prec)
 	return nil;
 }
 
+- (void)send_query:(dns_question_t *)q append:(LUArray *)a
+{
+	dns_reply_t *r;
+	LUArray *sub;
+	LUDictionary *item;
+	int i, len;
+
+	/* SDNS API */
+#ifdef DEBUG
+	log_query(q->name, q->class, q->type);
+#endif
+	r = dns_lookup(dns, q->name, q->dnsclass, q->dnstype);
+#ifdef DEBUG
+	log_reply(r);
+#endif
+	sub = [self arrayForDNSReply:r];
+
+	dns_free_reply(r);
+
+	if (sub == nil) return;
+
+	len = [sub count];
+	for (i = 0; i < len; i++)
+	{
+		item = [sub objectAtIndex:i];
+		[a addObject:item];
+	}
+
+	[sub release];
+}
+
 - (LUArray *)query:(LUDictionary *)pattern category:(LUCategory)cat
 {
-	LUArray *list, *all, *sub;
+	LUArray *list, *all;
 	LUDictionary *item;
-	dns_reply_t *r;
 	dns_question_t q;
 	char *name, *service, *protocol, *ip, *ipv6, *qname, *origname;
-	int fixup, i, j, len;
+	int fixup, i, len, srv;
 	int qtype[MAX_QTYPE], qtype_count;
 
 	if (cat != LUCategoryHost) return nil;
 	if (pattern == nil) return nil;
 	
-	q.class = DNS_CLASS_IN;	
+	q.dnsclass = ns_c_in;	
 
 	/*
 	 * search will be based on one of the following keys
@@ -1049,28 +1165,28 @@ precsize_ntoa(u_int8_t prec)
 	 *     ipv6_address
 	 */
 	qtype_count = 0;
-	qtype[qtype_count++] = DNS_TYPE_CNAME;
-
-	name = [pattern valueForKey:"name"];
-	if (name != NULL)
-	{
-		qtype[qtype_count++] = DNS_TYPE_A;
-		qtype[qtype_count++] = DNS_TYPE_AAAA;
-	}
+	srv = 0;
 
 	service = [pattern valueForKey:"service"];
 	protocol = [pattern valueForKey:"protocol"];
 	if ((service != NULL) || (protocol != NULL))
 	{
-		qtype[qtype_count++] = DNS_TYPE_SRV;
-		qtype[qtype_count++] = DNS_TYPE_MX;
+		qtype[qtype_count++] = ns_t_srv;
+		srv = 1;
+	}
+
+	name = [pattern valueForKey:"name"];
+	if ((srv == 0) && (name != NULL))
+	{
+		qtype[qtype_count++] = ns_t_a;
+		qtype[qtype_count++] = ns_t_aaaa;
 	}
 
 	ip = [pattern valueForKey:"ip_address"];
 	ipv6 = [pattern valueForKey:"ipv6_address"];
-	if ((ip != NULL) || (ipv6 != NULL))
+	if ((srv == 0) && ((ip != NULL) || (ipv6 != NULL)))
 	{
-		qtype[qtype_count++] = DNS_TYPE_PTR;
+		qtype[qtype_count++] = ns_t_ptr;
 	}
 
 	qname = NULL;
@@ -1085,11 +1201,10 @@ precsize_ntoa(u_int8_t prec)
 		{
 			if (protocol == NULL) return nil;
 			
-			qname = malloc(strlen(name) + strlen(service) + strlen(protocol) + 16);
 #ifdef RFC_2782
-			sprintf(qname, "_%s._%s.%s", service, protocol, name);
+			asprintf(&qname, "_%s._%s.%s", service, protocol, name);
 #else
-			sprintf(qname, "%s.%s.%s", service, protocol, name);
+			asprintf(&qname, "%s.%s.%s", service, protocol, name);
 #endif
 			fixup = FIX_SERV | FIX_PROT;
 			origname = copyString(name);
@@ -1122,30 +1237,28 @@ precsize_ntoa(u_int8_t prec)
 
 	for (i = 0; i < qtype_count; i++)
 	{
-		q.type = qtype[i];
+		q.dnstype = qtype[i];
+		[self send_query:&q append:all];
+	}
 
-		/* SDNS API */
-		r = sdns_query(dns, &q);
-		sub = [self arrayForDNSReply:r];
+	/* Special case for "smtp" - look up MX records */
+	if ((service != NULL) && (streq(service, "smtp")))
+	{
+		q.name = origname;
+		q.dnstype = ns_t_mx;
+		[self send_query:&q append:all];
+	}
 
-		dns_free_reply(r);
-
-		if (sub == nil) continue;
-
-		len = [sub count];
-		for (j = 0; j < len; j++)
-		{
-			item = [sub objectAtIndex:j];
-			[all addObject:item];
-		}
-
-		[sub release];
+	len = [all count];
+	if ((srv == 0) && (len == 0))
+	{
+		q.dnstype = ns_t_cname;
+		[self send_query:&q append:all];
+		len = [all count];
 	}
 
 	free(qname);
 
-	len = [all count];
-
 	for (i = 0; i < len; i++)
 	{
 		item = [all objectAtIndex:i];
@@ -1168,4 +1281,95 @@ precsize_ntoa(u_int8_t prec)
 	return list;
 }
 
+- (LUDictionary *)dns_proxy:(LUDictionary *)dict
+{
+	LUDictionary *item;
+	char *name, *cclass, *ctype, *proxy;
+	char buf[DNS_BUFFER_SIZE], b64[DNS_BASE_64_BUFFER_SIZE];
+	u_int16_t class, type;
+	int32_t test;
+	struct sockaddr *from;
+	u_int32_t do_search, fromlen, offset;
+	int32_t len;
+
+	name = [dict valueForKey:"name"];
+	if (name == NULL) return NO;
+
+	cclass = [dict valueForKey:"class"];
+	if (cclass == NULL) return NO;
+	class = atoi(cclass);
+
+	ctype = [dict valueForKey:"type"];
+	if (ctype == NULL) return NO;
+	type = atoi(ctype);
+
+	do_search = [dict unsignedLongForKey:"search"];
+	
+	fromlen = sizeof(struct sockaddr_storage);
+	from = (struct sockaddr *)calloc(1, fromlen);
+
+	len = -1;
+	if (do_search == 0)
+	{
+		len = dns_query(dns, name, class, type, buf, DNS_BUFFER_SIZE, from, &fromlen);
+	}
+	else
+	{
+		len = dns_search(dns, name, class, type, buf, DNS_BUFFER_SIZE, from, &fromlen);
+	}
+
+	if (len < 0)
+	{
+		free(from);
+		return NO;
+	}
+
+	proxy = [dict valueForKey:"proxy_id"];
+
+	item = [[LUDictionary alloc] initTimeStamped];
+	
+	test = b64_ntop(buf, len, b64, DNS_BASE_64_BUFFER_SIZE);
+
+	if (proxy != NULL) [item setValue:proxy forKey:"proxy_id"];
+	[item setValue:name forKey:"name"];
+	[item setValue:cclass forKey:"class"];
+	[item setValue:ctype forKey:"type"];
+	[item setValue:b64 forKey:"buffer"];
+
+	offset = 4;
+	if (from->sa_family == AF_INET6) offset = 8;
+	inet_ntop(from->sa_family, (char *)(from) + offset, buf, DNS_BUFFER_SIZE);
+	free(from);
+
+	[item setValue:buf forKey:"server"];
+
+#ifdef DEBUG
+	{
+		char d_buf[DNS_BUFFER_SIZE];
+		dns_reply_t *d_r;
+
+		test = b64_pton(b64, d_buf, DNS_BUFFER_SIZE);
+		if (test < 0) 
+		{
+			fprintf(stderr, "b64_pton failed!\n");
+		}
+		else
+		{
+			d_r = dns_parse_packet(d_buf, test);
+			if (d_r == NULL)
+			{
+				fprintf(stderr, "dns_parse_packet failed!\n");
+			}
+			else
+			{
+				dns_print_reply(d_r, stderr, 0xffff);
+				dns_free_reply(d_r);
+			}
+		}
+	}
+#endif
+
+	return item;
+}
+
 @end
